"""
Tools for conveying dslibrary calls to other systems.

* to_rest - translate calls into REST calls, securing them with an access token
* to_volume - communicate with a Kubernetes sidecar using a shared volume
* sample_rest - example rest service which converts http calls back into dslibrary calls
* volume_watcher - the other side of 'to_volume' which translates volume-based requests back into dslibrary calls
"""



"""
The 'other side' of communications for MMLibraryViaVolume.  Watches for requests appearing in a folder as JSON
files, and transmits them to a remote REST API.
"""
import base64
import os
import json
import time

from ..front import DSLibrary


class VolumeWatcher(object):
    def __init__(self, volume: str, target_dslibrary: DSLibrary):
        self._volume = volume
        self._target = target_dslibrary

    def scan_forever(self, callback=None, interval: float=0.05):
        """
        Scan repeatedly.  The supplied callback can raise an exception to stop the process.
        """
        try:
            while True:
                time.sleep(interval)
                if callback:
                    callback()
                self.scan_once()
        except StopIteration:
            pass

    def scan_once(self):
        """
        Scan for new requests.
        """
        for f in os.listdir(self._volume):
            if not f.endswith(".json"):
                continue
            try:
                fn = os.path.join(self._volume, f)
                with open(fn) as f_r:
                    request = json.load(f_r)
            except ValueError:
                continue
            response = self.process(request)
            # send response (if request file still exists)
            if os.path.exists(fn):
                response_fn = fn + ".response"
                response_tmp = fn + ".tmp"
                # write to temporary file
                with open(response_tmp, 'wb') as f_w:
                    if isinstance(response, (bytes, bytearray)):
                        f_w.write(response)
                    elif isinstance(response, str):
                        f_w.write(response.encode("utf-8"))
                    else:
                        resp_s = json.dumps(response)
                        f_w.write(resp_s.encode("utf-8"))
                # rename to expected location
                os.rename(response_tmp, response_fn)

    def process(self, request: dict):
        data = request.get("data")
        data_format = request.get("data_format")
        if data_format == "base64":
            data = base64.b64decode(data)
        # if the target has a communication method we can just forward everything directly
        if hasattr(self._target, "_do_comm"):
            return self._target._do_comm(
                method=request["method"], path=request["path"], params=request["params"],
                data=data, as_json=request["as_json"]
            )
        # otherwise we have to actually interpret the request
        path_parts = request["path"].split("/")
        method_path0 = f"{request['method']}:{path_parts[0]}"
        params = request.get("params", {})
        path = "/".join(path_parts[1:])
        # read file content
        if method_path0 in ("get:resources", "get:run_data"):
            if method_path0 == "get:run_data":
                opener = self._target.open_run_data
            else:
                opener = self._target.open_resource
            with opener(path, mode='rb') as f_r:
                if "byte_range" in params:
                    byte_range = json.loads(params["byte_range"])
                    f_r.seek(byte_range[0])
                    return f_r.read(byte_range[1] - byte_range[0])
                else:
                    return f_r.read()
        # write file content
        if method_path0 in ("put:resources", "put:run_data"):
            if method_path0 == "put:run_data":
                writer = self._target.write_run_data
            else:
                writer = self._target.write_resource
            append = params.get("append")
            writer(path, content=data, append=append)
            return {}
        # get metadata and parameters
        if method_path0 == "get:context":
            return {
                "metadata": self._target.get_metadata().to_json(),
                "parameters": self._target.get_parameters()
            }
        # scoring requests
        if method_path0 == "get:scoring-requests":
            request = self._target.next_scoring_request(timeout=params.get("timeout"))
            if request:
                return {"request": request}
            return {}
        if method_path0 == "post:score":
            self._target.scoring_response(params.get("score"))
            return {}
        # database r/w
        if method_path0 == "get:db":
            sql = params.get("sql")
            parameters = params.get("parameters")
            with self._target.get_sql_connection(path) as conn:
                c = conn.cursor()
                c.execute(sql, parameters)
                descr = c.description
                rows = list(c)
                more = False
            return [descr, rows, more]
        if method_path0 == "get:db_more":
            pass
        if method_path0 == "post:db":
            sql = params.get("sql")
            parameters = params.get("parameters")
            with self._target.get_sql_connection(path, for_write=True) as conn:
                c = conn.cursor()
                c.execute(sql, parameters)
            return {}
        # unrecognized
        raise Exception(f"not implemented yet: {method_path0}")



import os

from dslibrary import ENV_DSLIBRARY_SPEC, ENV_DSLIBRARY_TARGET


def reset_env(msg: str):
    if os.environ.get(ENV_DSLIBRARY_SPEC):
        print(f"WARNING: {ENV_DSLIBRARY_SPEC} was not cleared by prior tests ({msg})")
        os.environ[ENV_DSLIBRARY_SPEC] = ""
    if os.environ.get(ENV_DSLIBRARY_TARGET):
        print(f"WARNING: {ENV_DSLIBRARY_TARGET} was not cleared by prior tests ({msg})")
        os.environ[ENV_DSLIBRARY_TARGET] = ""




"""
A simple REST endpoint that just delegates to a local mmlibrary.  Demonstrates the server side of MMLibrary
endpoints.

(JUST GETTING STARTED WITH THIS)
"""
import json
import signal
import tornado.ioloop
import tornado.web
from .to_local import DSLibraryLocal


def detect_interrupt():
    request_close = []

    def signal_handler(signum, frame):
        request_close.append(1)

    def check_exit():
        if request_close:
            tornado.ioloop.IOLoop.instance().stop()

    signal.signal(signal.SIGINT, signal_handler)
    tornado.ioloop.PeriodicCallback(check_exit, 0.25*1000).start()


def create_app(target_folder: str):
    target = DSLibraryLocal(target_folder)

    class Context(tornado.web.RequestHandler):
        def get(self):
            ctx = {
                "uri": target.get_uri(),
                "parameters": target.get_parameters()
            }
            self.set_header("Content-Type", "application/json")
            self.write(json.dumps(ctx))

    class Resources(tornado.web.RequestHandler):
        def get(self):
            path = self.request.path.split("/resources/", maxsplit=1)[1]
            byte_range = json.loads(self.get_argument("byte_range"))
            with target.open_resource(path, mode='rb') as f_r:
                f_r.seek(byte_range[0])
                content = f_r.read(byte_range[1] - byte_range[0])
            self.set_header("Content-Type", "application/octet-stream")
            self.write(content)

        def put(self):
            path = self.request.path.split("/resources/", maxsplit=1)[1]
            append = self.get_argument("append") in ("true", 1, "True")
            hint = self.get_argument("hint")
            # TODO implement hints?  (for multi-part upload)
            with target.open_resource(path, mode='ab' if append else 'wb') as f_w:
                f_w.write(self.request.body)

    '''
    class AnotherOne(tornado.web.RequestHandler):
        async def get(self):
            try:
                value = json.loads(self.get_argument("request"))
            except ValueError:
                raise tornado.web.HTTPError(400, "Expected 'request' argument with JSON data")
            out = await ...
            self.set_header("Content-Type", "application/json")
            self.write(json.dumps({}))

        async def post(self):
            value = json.loads(self.request.body)
            out = await ba.request_one(self.request.path, value)
            self.set_header("Content-Type", "application/json")
            self.write(json.dumps({}))
    '''

    app = tornado.web.Application([
        (r"/context", Context),
        (r"/resources/.*", Resources),
    ])
    return app


def main():
    port = 6324
    app = create_app(".")
    app.listen(port)
    # tornado.ioloop.PeriodicCallback(ba.metrics.dump, settings.metrics_interval*1000).start()
    # tornado.ioloop.IOLoop.current().spawn_callback(ba.collect_batches)
    detect_interrupt()
    print(f"MMLibrary Endpoint running on port {port}")
    tornado.ioloop.IOLoop.current().start()


if __name__ == "__main__":
    main()
    exit(0)




import unittest

from dslibrary.utils.connect import _process_params


class TestConnect(unittest.TestCase):
    def test__process_params(self):
        r = _process_params("mysql://u:p@host:1234/database", extra=1)
        assert r == {
            'host': 'host',
            'port': 1234,
            'database': 'database',
            'username': 'u',
            'password': 'p',
            'extra': 1
        }, r
        r = _process_params("xyz://u@host")
        assert r == {
            'host': 'host',
            'port': None,
            'database': '',
            'username': 'u',
            'password': ''
        }, r
        r = _process_params("sqlite:path/to/file")
        assert r == {
            'host': '',
            'port': None,
            'database': 'path/to/file',
            'username': '',
            'password': ''
        }, r
        r = _process_params("x://host/db", database="db2")
        assert r == {
            'host': 'host',
            'port': None,
            'database': 'db2',
            'username': '',
            'password': ''
        }, r





"""
Remote, chunkable database connection: implements the standard Python API for a database connection,
calling remote services to read and write data.

Db interface specification: https://www.python.org/dev/peps/pep-0249/
"""
import time
import datetime
import re

SQL_PARTS = re.compile(r'(/\*.*?\*/|"([^"]|"")*?"|`[^`]*?`|\'([^\']|\'\')*?\'|((?!/\*)(?!%s)(?!\?)[^;\'"`])+|;|%s|\?)',
                       re.DOTALL)

apilevel = "1.0"
threadsafety = 2   # 1=module is thread-safe, 2=connections are thread-safe, 3=cursors are thread-safe
paramstyle = "format"  # expects '%s' for parameter replacements

Binary = bytes
Date = datetime.date
Time = datetime.time
Timestamp = datetime.datetime
DateFromTicks = lambda ticks: Date(*time.localtime(ticks)[:3])
TimeFromTicks = lambda ticks: Time(*time.localtime(ticks)[3:6])
TimestampFromTicks = lambda ticks: Timestamp(*time.localtime(ticks)[:6])


class _DBAPITypeObject:
    def __init__(self, *values):
        self.values = values
    def __eq__(self, other):
        if other in self.values:
            return True
        else:
            return False
    def __ne__(self, other):
        if other in self.values:
            return False
        else:
            return True

STRING = _DBAPITypeObject(str)
BINARY = _DBAPITypeObject(bytes)
NUMBER = _DBAPITypeObject(int, float, bool)
DATETIME = _DBAPITypeObject(datetime.datetime, datetime.date, datetime.time)
ROWID = _DBAPITypeObject()

class Error(Exception): pass
class Warning(Exception): pass
class InterfaceError(Error): pass
class DatabaseError(Error): pass
class InternalError(DatabaseError): pass
class OperationalError(DatabaseError): pass
class ProgrammingError(DatabaseError): pass
class IntegrityError(DatabaseError): pass
class DataError(DatabaseError): pass
class NotSupportedError(DatabaseError): pass


class Connection(object):
    def __init__(self, read: callable, write: callable=None, read_more: callable=None):
        """
        Create a DBI-compatible adapter that uses the supplied functions.
        :param read:        A method with two arguments: sql and parameters.  Returns cols, rows, more.  Cols will be
            returned for cursor 'description'.  Row is a list of row values.  More should be None if there are no more
            rows, or else a hint to read_more() for the next page.
        :param write:       A method with two arguments: sql and parameters.
        :param read_more:   A method with one argument: 'more'.  Returns cols, rows, more, (same as read()).
        """
        self.read = read
        self.read_more = read_more
        self.write = write

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def close(self):
        """ clean up """

    def commit(self):
        """ commit changes (placeholder) """

    def rollback(self):
        """ roll back changes (placeholder) """

    def cursor(self):
        return Cursor(self)


class Cursor(object):
    def __init__(self, conn: Connection):
        self._conn = conn
        self._cols = []
        self._rows = []
        self._pos = 0
        self.arraysize = 0
        self._ops = []

    @property
    def description(self):
        return self._cols

    @property
    def rowcount(self):
        return len(self._rows)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def callproc(self, procname, *parameters):
        raise NotSupportedError()

    def close(self):
        self._cols = []
        self._rows = []
        self._more = None
        self._ops = []
        self._pos = 0

    def execute(self, operation, parameters=None):
        self.close()
        tbl_only = _is_table_name(operation)
        if tbl_only:
            operation = 'select * from "%s"' % operation
        if _is_read_operation(operation):
            self._ops = list(sql_split(operation, parameters))
            self.nextset()
            return self
        else:
            if not self._conn.write:
                raise OperationalError("no write access")
            self._conn.write(operation, parameters)

    def nextset(self):
        if self._ops:
            op, op_params = self._ops.pop(0)
            self._cols, self._rows, self._more = self._conn.read(op, op_params)
            self._pos = 0
        else:
            raise OperationalError()

    def _ensure_next(self, n):
        # throw out rows as we go, but not too often to avoid wasting memory
        if self._pos > 1000:
            self._rows = self._rows[self._pos:]
            self._pos = 0
        # no 'read_more' function?
        if not self._conn.read_more:
            return
        # read more chunks
        while self._more:
            avail = len(self._rows) - self._pos
            if avail >= n:
                return
            more_rows, self._more = self._conn.read_more(self._more)
            self._rows += more_rows

    def fetchone(self):
        self._ensure_next(1)
        if self._pos < len(self._rows):
            p = self._pos
            self._pos += 1
            return self._rows[p]

    def fetchmany(self, size):
        self._ensure_next(size)
        avail = len(self._rows) - self._pos
        if size > avail:
            size = avail
        p = self._pos
        self._pos += size
        return self._rows[p: self._pos]

    def fetchall(self):
        self._ensure_next(1000000000000)
        p = self._pos
        self._pos = len(self._rows)
        return self._rows[p:]

    def executemany(self, operation, seq_of_parameters):
        if _is_read_operation(operation):
            raise OperationalError()
        else:
            for params in seq_of_parameters:
                self.execute(operation, params)

    def setinputsizes(self, sizes):
        """ placeholder """

    def setoutputsize(self, size, column=None):
        """ placeholder """

    def __next__(self):
        row = self.fetchone()
        if row:
            return row
        raise StopIteration

    def __iter__(self):
        return self


def _is_read_operation(operation):
    """
    Detect a read-only SQL operation.
    """
    verb = re.split(r'\s+', operation[:20].lower().strip())[0]
    return verb in {"select", "show", "describe"}


def _is_table_name(operation):
    """
    Detect "just a table name".  Returns the table name or None.
    """
    operation = operation.strip()
    m = re.match(r'^(([A-Za-z_][A-Za-z_0-9\-]*)|`([^`]+)`|"([^`]+)"|\[([^\]]+)\])$', operation)
    if m is None:
        return
    return m.group(2) or m.group(3) or m.group(4) or m.group(5)


def sql_split(operation, parameters=None):
    """
    Break apart SQL statements.  Returns a generator of SQL strings.
    """
    params_in = list(parameters or [])
    bits = []
    params = []
    for part in SQL_PARTS.finditer(operation):
        bit = part.group(0)
        if bit == ";":
            if bits:
                yield "".join(bits).strip(), params
            bits.clear()
            params = []
        else:
            if bit in {"?", "%s"}:
                params.append(params_in.pop(0) if params_in else None)
            bits.append(bit)
    if bits:
        yield "".join(bits).strip(), params


def sql_enquote_id(s: str, mode: str=None, always: bool=False, dotted: bool=False):
    """
    Enquote an identifier (table or column name) according to different conventions.
    :param s:       What to enquote.
    :param mode:    'mysql', 'postgres', 'ansi', ...
    :param always:  True to use quotes even when not required.
    :param dotted:  True when the name may contain multiple components, i.e. database.schema.table.  Otherwise dots are
                    considered part of the name to enquote.
    :return:    Identifier, quoted if necessary.
    """
    if dotted:
        return ".".join(map(lambda s1: sql_enquote_id(s1, mode=mode, always=always), s.split(".")))
    if s == "*" or s.startswith("*["):
        # special cases
        return s
    if re.match(r'^[a-zA-Z_][a-zA-Z0-9_\-]*$', s) is not None and not always:
        return s
    # we enquote file-based data sources like MySQL for an obscure reason:
    # - pandas.read_sql() strips double-quotes out of SQL!
    q = '`' if mode in {"mysql", "s3", "abs", "bigquery"} else '"'
    return "%s%s%s" % (q, s.replace(q, q+q), q)



"""
Abstraction of model execution.

Needs to work across all the normal targets:
  * Model Manager
  * mlflow
  * kubeflow, airflow
  * azure ml, sagemaker, gcp
  * kubernetes jobs
  * databricks

project / pipeline / model IDs
point to data sources
configurable engine
run a model
  parameters
  input data
  capture metrics, outputs
  errors
  tracking

this relates to:
  project enumeration
  model storage
"""
import typing


class Project(object):
    def __init__(self, uri: str=None, title: str=None):
        self.uri = uri
        self.title = title


class Pipeline(object):
    def __init__(self, uri: str=None, title: str=None):
        self.uri = uri
        self.title = title


class Model(object):
    def __init__(self, uri: str=None, title: str=None):
        self.uri = uri
        self.title = title


class Run(object):
    def __init__(self, uri: str=None, title: str=None):
        self.uri = uri
        self.title = title
        # TODO contain either a ModelRunner or a PipelineRunner


class ModelManagementEngine(object):
    """
    Top level view of a model management engine.
    """

    def list_projects(self, filters: dict=None) -> typing.Iterable[Project]:
        return []

    def list_pipelines(self, filters: dict=None) -> typing.Iterable[Pipeline]:
        return []

    def list_models(self, filters: dict=None) -> typing.Iterable[Model]:
        return []

    def list_runs(self, filters: dict=None) -> typing.Iterable[Run]:
        return []

    def list_schedules(self):
        # TODO define schedule descriptor
        pass

    # TODO ownership & permissions, esp. of projects & data sources
    # TODO environments
    # TODO configuration of data sources
    #   - when running a model we want to call open_resource("files1:/my_file.csv") and not care whether
    #     it is in s3, abs, or somewhere else
    #   - when calling get_db_connection("my_database") we shouldn't need to know whether it is BigQuery
    #     or PostGres, and what the connection string is
    #   - so....
    #       > we need to manage a set of data sources
    #       > we grant those data sources to models when we run them
    # TODO this is where a feature store would be integrated
    # TODO create/update/delete models & pipelines & data sources

    def model_runner(self, model_uri: str):
        """
        Launching a model begins here.
        """
        return ModelRunner(model_uri)

    def pipeline_runner(self, pipeline_uri: str):
        """
        Launching a pipeline begins here.
        """
        return PipelineRunner(pipeline_uri)


class ModelRunnerX(object):
    """

    """
    def __init__(self, model_uri: str):
        self._model_uri = model_uri
        self._arguments = {}
        self._run_id = None

    def set_argument(self, name: str, value):
        self._arguments[name] = value

    # TODO point to input data

    def start(self):
        # TODO return a tracking object
        pass

    def cancel(self):
        pass

    def wait(self, timeout: float=None):
        pass

    def results(self):
        # TODO return a read-only mmlibrary instance for accessing the inputs, arguments & results
        pass

    # TODO assign a run_id value
    # TODO branching result
    # TODO detect failures & get error detail, logs


class PipelineRunner(object):
    """
    Includes:
      * specification of the workflow (JSON)
      * execution of multiple models in a workflow
      * access to all the individual ModelRunner instances
    """
    def __init__(self, pipeline_uri: str):
        self._pipeline_uri = pipeline_uri

    def start(self):
        pass

    def cancel(self):
        pass



from urllib.parse import urlparse, unquote, urlunparse, quote
import re


def connect_to_database(uri: str, library: str=None, for_write: bool=False, **kwargs):
    """
    Try, by various means, to connect to a database.

    :param uri:        URI of database.
    :param library:     A specific library to use, if available
    :param for_write:   Whether to enable write operations.
    :returns:  a DBI driver
    """
    # determine flavor of connection and work out which libraries to try
    libraries = []
    if library:
        libraries.append(library)
    uri_parts = uri.split(":")
    if uri_parts[0] == "jdbc":
        uri_parts = uri_parts[1:]
    if uri_parts[0] in ("postgres", "postgresql"):
        libraries.append("psycopg2")
    if uri_parts[0] == "mysql":
        libraries.append("pymysql")
    if uri_parts[0] == "sqlite":
        libraries.append("sqlite")
    params = _process_params(uri, **kwargs)
    for lib in libraries:
        if lib in METHODS:
            conn = METHODS[lib](**params, for_write=for_write)
            if conn:
                return conn
    raise ValueError(f"No driver installed/supported for {uri_parts[0]}")


def _process_params(uri: str, **kwargs):
    """
    Use a URI to provide most of the needed values for connecting to a database.  'kwargs' supplies overrides.
    :param uri:         A URI which can supply most of the needed values.
    :param kwargs:      Additional overrides.
    :return:        A {} with all the extracted, named connection parameters.
    """
    if uri.startswith("jdbc:"):
        uri = uri[5:]
    if ":" not in uri:
        host, path = uri, ""
    else:
        parsed = urlparse(uri)
        host = parsed.netloc
        path = parsed.path
        if path.startswith("/"):
            path = path[1:]
    user = pwd = ""
    if "@" in host:
        # NOTE: '@' can occur in 'user', i.e. for Google's git repositories
        u_p, host = re.split(r'@(?!.*@)', host)
        if ":" in u_p:
            user, pwd = u_p.split(":")
        else:
            user, pwd = u_p, ""
    if ":" in host:
        host, port = host.split(":", maxsplit=1)
    else:
        port = None
    params = {
        "host": host,
        "port": int(port) if port else None,
        "database": path,
        "username": user,
        "password": pwd
    }
    params.update(kwargs)
    return params


def connect_pymysql(host, port=None, username=None, password=None, database=None, autocommit=True, for_write: bool=False, **kwargs):
    try:
        import pymysql
    except ImportError:
        return
    port = port or 3306
    conn = pymysql.connect(user=username, password=password, host=host, port=port, database=database or None, autocommit=autocommit, **kwargs)
    if not for_write:
        _set_conn_read_only(conn)
    return conn


def connect_psycopg2(host, port=None, username=None, password=None, database=None, autocommit=True, for_write: bool=False, **kwargs):
    try:
        import psycopg2
        import psycopg2.extensions
    except ImportError:
        return
    dsn_parts = {"host": host, "port": port or 5432, "dbname": database or "postgres"}
    if username:
        dsn_parts["user"] = username
    if password:
        dsn_parts["password"] = password
    pg_args_str = " ".join("%s=%s" % (k, v) for k, v in sorted(dsn_parts.items()))
    conn = psycopg2.connect(pg_args_str, **kwargs)
    if autocommit:
        conn.set_isolation_level(psycopg2.extensions.ISOLATION_LEVEL_AUTOCOMMIT)
    if not for_write:
        _set_conn_read_only(conn)
    return conn


def connect_sqlite(host, database=None, for_write: bool=False, **kwargs):
    try:
        import sqlite3
    except ImportError:
        return
    for ignore in ["port", "username", "password", "database", "autocommit"]:
        kwargs.pop(ignore, None)
    if not for_write:
        return sqlite3.connect(f'file:{host or database}?mode=ro', uri=True, **kwargs)
    return sqlite3.connect(host, **kwargs)


def _set_conn_read_only(conn):
    """
    Try to enforce a read-only connection
    """
    try:
        conn.cursor().execute("set transaction read only")
    except:
        print(f"couldn't set read-only mode on connection: {conn.__class__.__name__}")


METHODS = {
    "pymysql": connect_pymysql,
    "psycopg2": connect_psycopg2,
    "sqlite": connect_sqlite
}

# TODO BigQuery



import unittest
import mock
import io
import os
import numpy
import pandas
import shutil
import tempfile

from dslibrary import DSLibrary
from dslibrary.front import EVALUATION_RESULT_ALIAS, METRICS_ALIAS
from dslibrary.metadata import Metadata
from dslibrary.transport.to_local import DSLibraryLocal


class TestFront(unittest.TestCase):

    def test_metadata(self):
        """
        The base class doesn't know how to get metadata so it returns a null instance.
        """
        dsl = DSLibrary()
        m = dsl.get_metadata()
        assert isinstance(m, Metadata)
        assert m.uri == ""
        assert m.entry_points == {}

    def test_open_resource__input_mapping(self):
        """
        Options for input come from mapping values, which override supplied values.
        """
        log = []
        class MyInst(DSLibrary):
            def _opener(self, path: str, mode: str, **kwargs):
                log.append((path, mode, kwargs))
                return "H"
        MyInst(spec={"inputs": {"a": {"uri": "aaa", "option1": 1, "option3": 3}}}).open_resource("a", option1=2, option2=22)
        assert log[0] == ('aaa', 'rb', {'option1': 1, 'option2': 22, 'option3': 3}), log[0]

    def test_open_resource__output_mapping(self):
        """
        Options for output come from mapping values, which override supplied values.
        """
        log = []
        class MyInst(DSLibrary):
            def _opener(self, path: str, mode: str, **kwargs):
                log.append((path, mode, kwargs))
                return "H"
        MyInst(spec={"outputs": {"a": {"uri": "aaa", "option1": 1, "option3": 3}}}).open_resource("a", 'w', option1=2, option2=22)
        self.assertEqual(log[0], ('aaa', 'w', {'option1': 1, 'option2': 22, 'option3': 3}))

    def test_open_resource__bypass_mapping(self):
        """
        Most inputs and outputs are mapped, but you can bypass mapping and specify a URI to open or a local file.
        """
        log = []
        class MyInst(DSLibrary):
            def _opener(self, path: str, mode: str, **kwargs):
                log.append((path, mode, kwargs))
                return "H"
        MyInst().open_resource("s3://bucket/path", open_option_1=1)
        assert log[0] == ('s3://bucket/path', 'rb', {'open_option_1': 1})
        MyInst().open_resource("./local/file")
        assert log[1] == ('./local/file', 'rb', {})

    def test_open_model_binary(self):
        """
        The default implementation just assumes certain filenames for the 'model-binary' data.
        """
        log = []
        class MyInst(DSLibrary):
            def _opener(self, path: str, mode: str, **kwargs):
                log.append((path, mode, kwargs))
                return "H"
        MyInst().open_model_binary()
        assert log[0] == ('model-binary', 'rb', {})
        MyInst().open_model_binary("part1")
        assert log[1] == ('model-binary/part1', 'rb', {})

    def test_set_evaluation_result(self):
        """
        Model evaluation code can use this hook to report whether the model passed or failed.
        """
        log = []
        class MyInst(DSLibrary):
            def _opener(self, path: str, mode: str, **kwargs):
                buf = io.StringIO()
                buf.close = lambda: None
                log.append((path, buf))
                return buf
        MyInst().set_evaluation_result(True)
        MyInst().set_evaluation_result(False, reason="because")
        assert log[0][0] == log[1][0] == EVALUATION_RESULT_ALIAS
        r = log[0][1].getvalue()
        assert r == '{"uri": "", "success": true}\n', r
        r = log[1][1].getvalue()
        assert r == '{"uri": "", "success": false, "reason": "because"}\n', r

    def test_get_sql_connection__mapping(self):
        """
        The normal use case is have the caller supply all the connection information.
        """
        dsl = DSLibrary(spec={"inputs": {"db": {"uri": "mysql://host/db", "username": "u"}}})
        def connect(user, host, database, **kwargs):
            assert user == "u"
            assert host == "host"
            assert database == "db"
            return "CONN"
        with mock.patch("pymysql.connect", connect):
            r = dsl.get_sql_connection("db")
            assert r == "CONN"

    def test_write_resources__dataframes_and_series(self):
        """
        Several types are supported for columnar data.
        """
        log = []
        class MyInst(DSLibrary):
            def _opener(self, path: str, mode: str, **kwargs):
                buf = io.StringIO()
                buf.close = lambda: None
                log.append((path, buf))
                return buf
        dsl = MyInst()
        # series
        dsl.write_resource("x", pandas.Series([1, 2, 3]))
        assert log[0][0] == "x"
        self.assertEqual(log[0][1].getvalue(), 'x\n1\n2\n3\n')
        log.clear()
        # numpy array
        dsl.write_resource("x", numpy.array([1, 2, 3]))
        assert log[0][0] == "x"
        self.assertEqual(log[0][1].getvalue(), 'x\n1\n2\n3\n')
        log.clear()
        # dataframe
        dsl.write_resource("x", pandas.DataFrame({"y": [1, 2, 3]}))
        assert log[0][0] == "x"
        self.assertEqual(log[0][1].getvalue(), 'y\n1\n2\n3\n')
        log.clear()

    def test_default_metrics_output(self):
        project = tempfile.mkdtemp()
        dsl = DSLibraryLocal(project)
        dsl.log_metric("x", 1)
        r = dsl.get_last_metric("x")
        assert r.value == 1
        # verify it was written as JSON
        fn = os.path.join(project, METRICS_ALIAS)
        assert os.path.exists(fn)
        with open(fn, 'r') as f_r:
            assert f_r.read().startswith("{")
        shutil.rmtree(project)

    def test_alt_metrics_output1(self):
        project = tempfile.mkdtemp()
        dsl = DSLibraryLocal(project, spec={"outputs": {METRICS_ALIAS: {"format": "csv"}}})
        dsl.log_metric("x", 1)
        # verify it was written as CSV
        fn = os.path.join(project, METRICS_ALIAS)
        assert os.path.exists(fn)
        with open(fn, 'r') as f_r:
            assert f_r.read().startswith("uri,")
        shutil.rmtree(project)

    def test_alt_metrics_output2(self):
        project = tempfile.mkdtemp()
        dsl = DSLibraryLocal(project, spec={"outputs": {METRICS_ALIAS: {"uri": "metrics.csv"}}})
        dsl.log_metric("x", 1)
        # verify it was written as CSV
        fn = os.path.join(project, "metrics.csv")
        assert os.path.exists(fn)
        with open(fn, 'r') as f_r:
            assert f_r.read().startswith("uri,")
        shutil.rmtree(project)

    def test_model_pickling(self):
        project = tempfile.mkdtemp()
        dsl = DSLibraryLocal(project, spec={"outputs": {METRICS_ALIAS: {"uri": "metrics.csv"}}})
        # save and restore
        my_model = {"x": 1}
        dsl.save_pickled_model(my_model)
        restored = dsl.load_pickled_model()
        assert restored == my_model
        # verify local storage
        assert os.path.exists(project + "/model-binary")
        shutil.rmtree(project)

    def test_open_resource__default_uri(self):
        """
        Specify a name and a default URI, for cleaner overriding.
        """
        log = []
        class MyInst(DSLibrary):
            def _opener(self, path: str, mode: str, **kwargs):
                log.append((path, kwargs))
                return "H"
        MyInst().open_resource("x", uri="default")
        MyInst(spec={"inputs": {"x": {"uri": "override", "option1": 1}}}).open_resource("x", uri="default")
        self.assertEqual(log, [('default', {}), ('override', {'option1': 1})])

    def test_load_dataframe__default_uri(self):
        log = []
        class MyInst(DSLibrary):
            def _opener(self, path: str, mode: str, **kwargs):
                log.append(path)
                return io.BytesIO(b"x\n1\n2")
        df = MyInst().load_dataframe("x", uri="default.csv")
        df = MyInst(spec={"inputs": {"x": {"uri": "override.csv"}}}).load_dataframe("x", uri="default.csv")
        self.assertEqual(log, ['default.csv', 'override.csv'])

    def test_load_dataframe__from_sql(self):
        log = []
        class MyInst(DSLibrary):
            def get_sql_connection(self, resource_name: str, for_write: bool=False, **kwargs):
                class DbCursor(object):
                    description = [("x", None)]
                    def execute(self, sql):
                        log.append(sql)
                    def __iter__(self):
                        return iter([(1,), (2,)])
                class DbConn(object):
                    def cursor(self):
                        return DbCursor()
                    def close(self):
                        log.append("close")
                return DbConn()
        df = MyInst().load_dataframe("x", sql_table="table1")
        self.assertEqual(list(df.x), [1, 2])
        self.assertEqual(log, ['SELECT * from table1', 'close'])

    def test_load_dataframe__from_nosql(self):
        class MyInst(DSLibrary):
            def get_nosql_connection(self, resource_name: str, for_write: bool=False, **kwargs):
                class MyNoSql(object):
                    def query(self, collection, **kwargs):
                        assert collection == "table1"
                        return [{"x": 1}, {"x": 2}]
                return MyNoSql()
        df = MyInst().load_dataframe("x", nosql_collection="table1")
        self.assertEqual(list(df.x), [1, 2])

    def test_load_dataframe__sql__custom_open_args(self):
        """
        Custom arguments can be passed through to get_sql_connection().
        """
        class MyInst(DSLibrary):
            def get_sql_connection(self, resource_name: str, database: str=None, for_write: bool=False, **kwargs):
                assert resource_name == "engine"
                assert database == "db"
                assert kwargs["custom1"] == 123
                class DbCursor(object):
                    description = [("x", None)]
                    def execute(self, sql):
                        pass
                    def __iter__(self):
                        return iter([])
                class DbConn(object):
                    def cursor(self):
                        return DbCursor()
                    def close(self):
                        pass
                return DbConn()
        MyInst().load_dataframe("engine", sql_table="table1", database="db", custom1=123)

    def test_open_resource__named_filesystem(self):
        """
        A resource can be opened through a specified filesystem engine/provider.
        """
        class MyInst(DSLibrary):
            def get_filesystem_connection(self, resource_name: str, for_write: bool=False, **kwargs):
                assert resource_name == "fs"
                assert kwargs.get("custom1") == 111
                class FS(object):
                    def open(self, path, mode):
                        assert path == "path"
                        assert mode == "r"
                        return "FH"
                return FS()
        r = MyInst().open_resource("path", mode="r", filesystem="fs", custom1=111)
        assert r == "FH"

    def test_load_dataframe__named_filesystem(self):
        """
        A dataframe can be opened through a specified filesystem engine/provider.
        """
        class MyInst(DSLibrary):
            def get_filesystem_connection(self, resource_name: str, for_write: bool=False, **kwargs):
                assert resource_name == "fs"
                class FS(object):
                    def open(self, path, mode):
                        assert path == "path"
                        assert mode == "rb"
                        return io.BytesIO(b"x\n1\n2\n3")
                return FS()
        df = MyInst().load_dataframe("path", filesystem="fs", format="csv")
        assert list(df.x) == [1, 2, 3]



"""
Implementation of calls for local development, experimentation and testing.
"""
import json
import os
import time

from dslibrary.utils.file_utils import FileOpener
from dslibrary.front import DSLibrary, DSLibraryException
from dslibrary.metadata import Metadata


class DSLibraryLocal(DSLibrary):
    """
    Functionality is configured through 'spec'.
    """
    def __init__(self, filesystem_root: str=None, spec: (dict, str)=None, _env: dict=None):
        super(DSLibraryLocal, self).__init__(spec, _env=_env)
        self._root = os.path.expanduser(filesystem_root or ".")
        self._metadata = None
        self._params = None

    def get_metadata(self):
        """
        Information about parameters, inputs, outputs.  We check for a few specific filenames in the model's root
        folder.
        """
        if self._metadata is None:
            self._metadata = Metadata.from_folder(self._root, fill_default=True)
        return self._metadata

    def _opener(self, path: str, mode: str, **kwargs):
        """
        Open files.

        :param path:    Path to file, or URI of file.
        :param mode:    Open mode (r, rb, w, wb, a, ab)
        :param kwargs:  Additional arguments to customize details of the operation.
        :return:    File-like object.
        """
        # TODO  this is a workaround for a bug in "adlfs".  fsspec.open("abfs://...", ...) does not work: it eventually gets a stack overflow
        if path.startswith("abfs://"):
            fs = self.get_filesystem_connection(path, for_write="r" not in mode, **kwargs)
            path_only = path.split("://")[1].split("/", maxsplit=1)[1]
            return fs.open(path_only, mode=mode)
        return FileOpener(self._root).open(path, mode, **kwargs)

    def open_run_data(self, filename: str, mode: str = 'r'):
        """
        We put all the 'run data' in a sibling folder, shared across all related models.
        """
        if self._mlflow_all:
            raise DSLibraryException("open_run_data() is not yet supported for use in MLFlow")
        tmp_folder = os.path.join(self._root, "../__run_data__")
        if not os.path.exists(tmp_folder) and ('w' in mode or 'a' in mode):
            os.mkdir(tmp_folder)
        return self._opener(os.path.join(tmp_folder, filename.strip("/")), mode)

    def scoring_response(self, score) -> None:
        """
        Scores are appended to a CSV file.
        """
        self.write_resource("scores.json", json.dumps({"time": time.time(), "score": score}) + "\n", append=True)

    def iter_scoring_requests(self, callback=None):
        """
        This implementation supports a very simple way to test a scoring engine.  You drop JSON files into a
        folder to make a request.  When the folder is deleted, iteration ends and the scoring engine exits.
        """
        folder = os.path.join(self._root, "scoring_requests")
        while True:
            if callback:
                callback()
            time.sleep(0.2)
            if not os.path.exists(folder):
                continue
            for f in os.listdir(folder):
                if not f.endswith(".json"):
                    continue
                fn = os.path.join(folder, f)
                try:
                    with open(fn, 'r') as f_r:
                        request = json.load(f_r)
                except ValueError:
                    # ignore invalid format (file may not be completely written)
                    # - note that invalid files will stack up
                    continue
                os.remove(fn)
                yield request



import unittest
import tempfile
import shutil

from dslibrary.utils.file_utils import FileOpener, connect_to_filesystem, write_stream_with_read_on_close, \
    find_url_extension


class TestFileUtils(unittest.TestCase):

    def test_FileOpener_local_file(self):
        tmp_f = tempfile.mkdtemp()
        fo = FileOpener(tmp_f)
        with fo.open("x", mode="w") as f_w:
            f_w.write("abc")
        with fo.open("x", mode="r") as f_r:
            self.assertEqual(f_r.read(), "abc")
        shutil.rmtree(tmp_f)

    def test_write_stream_with_read_on_close(self):
        log = []
        f_w = write_stream_with_read_on_close('w', 'r', on_close=lambda fh: log.append(fh.read()))
        f_w.write("abc")
        f_w.write("def")
        f_w.close()
        assert log == ["abcdef"]

    def test_connect_to_filesystem(self):
        tmp_f = tempfile.mkdtemp()
        fs = connect_to_filesystem(tmp_f, for_write=True)
        with fs.open("x", mode="w") as f_w:
            f_w.write("abc")
        with fs.open("x", mode="r") as f_r:
            self.assertEqual(f_r.read(), "abc")
        self.assertEqual(fs.ls(), [{'name': 'x', 'size': 3, 'type': 'file'}])
        self.assertEqual(fs.stat("x"), {'name': 'x', 'size': 3, 'type': 'file'})
        assert fs.exists("x") is True
        assert fs.exists("y") is False
        shutil.rmtree(tmp_f)

    def test_find_url_extension(self):
        self.assertEqual(find_url_extension("abc.ext"), "ext")
        self.assertEqual(find_url_extension("abc.Ext"), "ext")
        self.assertEqual(find_url_extension("abc"), "")
        self.assertEqual(find_url_extension("http://host/path/filename.ext"), "ext")
        self.assertEqual(find_url_extension("abc.ext?option=1"), "ext")
        self.assertEqual(find_url_extension("abc.ext#hash_etc"), "ext")



"""
Various utilities.
"""
import importlib
import inspect


def valueize(s: str):
    try:
        return float(s)
    except:
        return s


def parse_verb_and_args(verb_and_args: str):
    """
    Parse a general purpose specification consisting of a verb, positional arguments, and named arguments.
    :param verb_and_args:   VERB:POSITIONAL_ARGUMENT:NAME=NAMED_ARGUMENT
    :return:   A tuple with the verb, args and kwargs.
    """
    parts = verb_and_args.split(":")
    args = []
    kwargs = {}
    for arg in parts[1:]:
        arg_parts = arg.split('=', maxsplit=1)
        if len(arg_parts) == 1:
            args.append(valueize(arg))
        else:
            kwargs[arg_parts[0]] = valueize(arg_parts[1])
    return parts[0], tuple(args), kwargs


def instantiate_class_and_args(class_and_args: str, required_type=None):
    """
    Instantiate a class of a given type, given a string specifying a class name and constructor arguments.

    See parse_verb_and_args() for format explanation.

    :param class_and_args:      PACKAGE.CLASS:ARG:NAME=ARG
    :param required_type:       A type, or list of types which must be matched.
    :return:                An instance of the specified class.
    """
    cls_name, args, kwargs = parse_verb_and_args(class_and_args)
    try:
        cls = import_named_object(cls_name)
        if required_type and not issubclass(cls, required_type):
            raise ValueError(f"Specified class {cls_name} is not {required_type}")
        return cls(*args, **kwargs)
    except ImportError:
        raise ValueError(f"Specified class {cls_name} was not found")


def import_named_object(package_and_object: str):
    """
    Import a class or method based on a string giving a package, module and attribute name delimited by '.'.
    :param package_and_object:   PACKAGE.MODULE.NAME
    :return:    Method imported from the indicated module.
    """
    parts = package_and_object.split(".")
    module = importlib.import_module('.'.join(parts[:-1]))
    return getattr(module, parts[-1])



import unittest
import io

from dslibrary import DSLibrary
from dslibrary.transport.http_to_dsl import HttpToDsl, RequestInfo


class TestHttpToDsl(unittest.TestCase):

    def test_rw_resource(self):
        files = {"a": io.BytesIO(b'abc')}
        class MyDsl(DSLibrary):
            def open_resource(self, resource_name: str, mode: str='rb', **kwargs):
                if mode == 'wb':
                    stream = io.BytesIO()
                    stream.close = lambda: None
                    files[resource_name] = stream
                    return stream
                if mode == 'rb':
                    return files[resource_name]
        xlt = HttpToDsl(MyDsl(), base_path="api")
        r = xlt.translate(RequestInfo("get", "/api/resources?resource_name=a"))
        assert r.content == b'abc'
        r = xlt.translate(RequestInfo("put", "/api/resources?resource_name=a", post_data=b'def'))
        assert r.content == {}
        assert files["a"].getvalue() == b'def'



"""
Using to_volume and volume_watcher.
"""
import unittest
import tempfile
import threading
import shutil

from dslibrary.transport.to_local import DSLibraryLocal
from dslibrary.transport.to_volume import DSLibraryViaVolume
from dslibrary.transport.volume_watcher import VolumeWatcher


class TestVolumeComm(unittest.TestCase):

    def vol_to_vol_setup(self):
        class Setup(object):
            def __enter__(self):
                self.volume = tempfile.mkdtemp()
                self.target_data = tempfile.mkdtemp()
                self.front = DSLibraryViaVolume(self.volume)
                self.target = DSLibraryLocal(self.target_data)
                self.xfer = VolumeWatcher(self.volume, self.target)
                # bg thread to manage 'xfer'
                self.done = []
                def my_callback():
                    if self.done:
                        raise StopIteration
                self.bg = threading.Thread(target=lambda: self.xfer.scan_forever(callback=my_callback))
                self.bg.start()
                return self

            def __exit__(self, exc_type, exc_val, exc_tb):
                self.done.append(1)
                shutil.rmtree(self.volume)
                shutil.rmtree(self.target_data)
                self.bg.join()
        return Setup()

    def test_file_rw(self):
        with self.vol_to_vol_setup() as env:
            # read data from target
            env.target.write_resource("file1", "DATA1")
            self.assertEqual(env.front.read_resource("file1", mode='r'), "DATA1")
            self.assertEqual(env.front.read_resource("file1", mode='rb'), b"DATA1")
            env.target.write_resource("file1", b"DATA1")
            self.assertEqual(env.front.read_resource("file1", mode='r'), "DATA1")
            self.assertEqual(env.front.read_resource("file1", mode='rb'), b"DATA1")
            # write data to target
            env.front.write_resource("file2", "DATA2")
            self.assertEqual(env.target.read_resource("file2", 'r'), "DATA2")
            self.assertEqual(env.target.read_resource("file2", 'rb'), b"DATA2")
            env.front.write_resource("file2", b"DATA2")
            self.assertEqual(env.target.read_resource("file2", 'r'), "DATA2")
            self.assertEqual(env.target.read_resource("file2", 'rb'), b"DATA2")



"""
The model-facing side of this API.
"""
import io
import json
import jsonschema
import os
import pickle
import uuid
import sys
import pandas
import numpy
import time
import typing
from collections import namedtuple
import yaml

from .proto_handlers import load_protocol_handler
from .utils.connect import connect_to_database
from .utils.dbconn import sql_enquote_id
from .utils.file_utils import write_stream_with_read_on_close, connect_to_filesystem, find_url_extension
from .utils.nosql import connect_to_nosql
from .utils.packages import install_packages

try:
    import mlflow
except ImportError:
    mlflow = None

from .metadata import Metadata

# this environment variable says how DSLibrary should be implemented
ENV_DSLIBRARY_TARGET = "DSLIBRARY_TARGET"
# this one provides a JSON specification indicating all the parameters and where the data should be channeled
ENV_DSLIBRARY_SPEC = "DSLIBRARY_SPEC"
# this one is used to provide authentication when communicating with a remote REST service
ENV_DSLIBRARY_TOKEN = "DSLIBRARY_TOKEN"

# use this as the name of an input or output to access metrics
METRICS_ALIAS = "__metrics__"
# use this to specify where parameters are logged
PARAMS_ALIAS = "__params__"
# pass/fail signal from model evaluation actions
EVALUATION_RESULT_ALIAS = "__evaluation_result__"


class DSLibrary(object):
    """
    Base class for all dslibrary implementations.
    """
    # internal representation of a stored metric
    Metric = namedtuple("Metric", "run_id uri user time name value step")

    def __init__(self, spec: dict=None, _env: dict=None):
        """
        :param spec:    See RunModel for a description of (and a generator for) the specification.  It provides
                        parameter values, and says where input and output data should go, what file format to use, and
                        other details of the context in which the model is operating, like whether to delegate any
                        methods to MLFlow.
        """
        self._params = None
        self._spec = spec or {}
        self._env = _env
        spec_wo_data = dict(self._spec)
        spec_wo_data.pop("data", None)
        jsonschema.validate(spec_wo_data, LOCAL_SPECS_SCHEMA)
        self._run_id = self._spec.get("run_id") or ""
        self._mlflow_opts = self._spec.get("mlflow") or {}
        self._mlflow_all = self._mlflow_opts.get("all")
        self._mlflow_metrics = self._mlflow_opts.get("metrics") or self._mlflow_all
        if self._mlflow_opts and not mlflow:
            raise DSLibraryException("MLFlow was selected for output but the mlflow package is not installed")
        self._mlflow_run = None

    def _xlt_resource(self, resource_name: str, _mode: str="inputs", **kwargs):
        """
        Look up specifications for a named input or output and use those to override supplied defaults.
        :param resource_name:   Named input.
        :param kwargs:          Defaults.
        :return:    A tuple with (0) the URI, and (1) the final set of options
        """
        resource_spec = (self._spec.get(_mode) or {}).get(resource_name) or {}
        kwargs.update(resource_spec)
        uri = kwargs.pop("uri", None) or resource_name
        return uri, kwargs

    def get_metadata(self) -> Metadata:
        """
        Load self-descriptive metadata for this model.  Contains information about parameters, inputs, & outputs.
        Similar to (and aspirationally compatible with) content of mlflow's MLProject file.

        This information would normally come from the model code package itself, but subclasses are free to source it
        from wherever they like.

        The main points covered by the metadata are:
          * entry points - a model can have multiple actions that it can perform, such as training or prediction, and
                most models have a default ('main') action.  Each entry point is associated with a command that can be
                executed, generally involving some source code that is part of the model.
          * parameter names and types - each entry point may have parameters, and the model is made more usable when
                the names and types of those parameters are known without digging into the code.  They can be validated,
                defaults can be filled in, and so on.
          * input and output schemas - each entry point will generally read or write columnar data, and that data will
                usually have some rules, like required columns that must be present, or expected column types.
                Documenting these allows the surrounding infrastructure to ensure that these rules are followed.
        """
        return Metadata()

    def get_parameters(self):
        """
        Parameter values come from supplied specifications (see constructor), or if omitted, we fall back to the command
        line.
        """
        if self._params is None:
            # check local specifications, then check environment and command line
            params = self._spec.get("parameters")
            # fall back to CLI parameters
            if params is None:
                params = self._scan_cli_parameters()
            # fill in defaults, perform type coercion, do additional validation
            metadata = self.get_metadata()
            entry_point = metadata.entry_points.get(self._spec.get("entry_point") or "main")
            if entry_point:
                for param_name, param_props in entry_point.parameters.items():
                    params[param_name] = param_props.process_value(params.get(param_name))
            # cache the parameter values
            self._params = params
        return self._params

    def _scan_cli_parameters(self) -> dict:
        """
        Load parameters from the command line.
        """
        args = {}
        param_next = None
        for arg in sys.argv:
            if param_next:
                v = arg
                if v.isdigit():
                    v = float(v)
                args[param_next] = v
                param_next = None
            elif arg.startswith("--") and arg[2:]:
                if '=' in arg:
                    k, v = arg[2:].split('=', maxsplit=1)
                    if v.isdigit():
                        v = float(v)
                    args[k] = v
                else:
                    param_next = arg[2:]
        return args

    def get_parameter(self, parameter_name: str, default=None):
        """
        Get a defined parameter.  Returns (default) if not defined.
        """
        return self.get_parameters().get(parameter_name, default)

    def get_uri(self) -> str:
        """
        Returns a URI that identifies the model which is currently running.  This is an identifier for a particular
        model or piece of executable code, not for a particular execution of the code.
        """
        # TODO this is given in the metadata but can also be specified in the specification, which is confusing.
        #  - MLFlow assumes it can access a model's project data based on a URI, which makes it seem awkward to expect
        #    it in the metadata file.
        return self.get_metadata().uri or self._spec.get("uri") or ""

    def _opener(self, path: str, mode: str, **kwargs) -> io.RawIOBase:
        """
        Open files.  Derived classes fill this in with a means of opening read or write streams.  It should ensure the
        filename is within allowed bounds for the model, i.e. "../" and "/..." are only allowed in some cases.

        :param path:    Path to file, or URI of file.
        :param mode:    Open mode (r, rb, w, wb, a, ab)
        :param kwargs:  Additional arguments to customize details of the operation.
        :return:    File-like object.  FileNotFoundError if not found for read.
        """
        raise DSLibraryException("base class called")

    def open_resource(self, resource_name: str, mode: str='rb', **kwargs) -> io.RawIOBase:
        """
        Open a stream to or from a named file-like source/destination, or from a URI.  Each input and output should be
        given a name.  The caller selects particular data sources and associates them with these names.  The burden of
        indicating file format details is on the caller.

        A good practice is to specify a logical name in 'resource_name', and provide a default uri in 'uri'.

        NOTES:
          * Inputs and outputs may be segregated: writing to an output named 'x' may have no effect on the input
            named 'x'.
          * Not all file sources support append mode (i.e. s3 buckets).

        :param resource_name:   The path or URI indicating which resource to access.
            * Names starting with './' are treated as local files.
            * Names starting with 'http://', 'https://', 's3://', etc., are treated as URIs of external resources.
            * All other names are expected to match an input or output defined in the metadata.

        :param mode:            File read/write/append mode: 'r', 'rb', 'w', 'wb', 'a', 'ab'

        :param kwargs:      Additional arguments:
                                filesystem = opens the file through a named external filesystem engine

        :returns:       A file-like object.
        """
        uri, open_args = self._xlt_resource(resource_name, _mode="inputs" if "r" in mode else "outputs", **kwargs)
        # remove formatting arguments
        open_args.pop("format", None)
        open_args.pop("format_options", None)
        # send all outputs to mlflow if requested
        use_mlflow = self._mlflow_all and ("w" in mode or "a" in mode)
        if use_mlflow:
            if "a" in mode:
                raise DSLibraryException("Append mode not available with mlflow output")
            return write_stream_with_read_on_close(w_mode=mode, r_mode='r', on_close=lambda fh: mlflow.log_artifact(fh.name, uri))
        # check for a custom protocol handler
        proto_handler = load_protocol_handler(uri, env=self._env)
        if proto_handler:
            return proto_handler.open_resource(uri, mode, **open_args)
        # data can be requested through a named filesystem engine/provider
        filesystem = kwargs.get("filesystem")
        if filesystem:
            fs = self.get_filesystem_connection(filesystem, for_write="r" not in mode, **open_args)
            # TODO some of 'open_args' might be appropriate to send to open() below rather than get_filesystem_connection(), above
            return fs.open(uri, mode=mode)
        # 'path' has the URI or path of what we need to open
        # 'mode' has the open mode
        # 'open_args' has additional arguments to pass
        return self._opener(uri, mode, **open_args)

    def open_model_binary(self, part: str=None, mode: str='rb') -> io.RawIOBase:
        """
        Read or write to the main model data store for the current model.  If there is just one piece of data, like
        a pickled sklearn model, omit the 'part' argument.  If there are multiple parts, 'part' indicates which one
        to access.

        Reading from this file would be the normal operation when making a prediction.  Writing to this file implies
        a training operation.
        """
        return self.open_resource(f"model-binary/{part}" if part else "model-binary", mode=mode)

    def open_run_data(self, filename: str, mode: str='rb') -> io.RawIOBase:
        """
        Open a stream to or from file-like storage which is local to the current pipeline execution context.

        One model can write to a 'file' in this context, and another can read from that file, provided they are part
        of the same pipeline.
        """
        raise DSLibraryException("open_run_data() is not supported by the selected implementation")

    def set_evaluation_result(self, success: bool, **kwargs) -> None:
        """
        Signal the completion of an evaluation operation, indicating whether it succeeded, and supplying an optional
        message.
        """
        self.log_dict({"uri": self.get_uri(), "success": success, **kwargs}, EVALUATION_RESULT_ALIAS)

    def _get_system_connection(self, system_type, resource_name: str, for_write: bool=False, **kwargs):
        """
        Access all types of external system connections.
        :param system_type:         See methods below.
        :param resource_name:       Named input or output, or URI.
        :param for_write:           Request write access.
        :param kwargs:              Custom arguments.
        :return:        An engine implementation, the type of which depends on the requested system_type.
                        See 'engine_intf'.
        """
        # translate 'resource_name' based on supplied specification and overriding supplied arguments
        uri, kwargs = self._xlt_resource(resource_name, "outputs" if for_write else "inputs", **kwargs)
        handler = load_protocol_handler(uri, env=self._env)
        if handler:
            return handler.get_system_connection(system_type, uri=resource_name, for_write=for_write, **kwargs)
        if system_type == "filesystem":
            return connect_to_filesystem(uri=uri, for_write=for_write, **kwargs)
        if system_type == "sql":
            library = kwargs.pop("library", None)
            return connect_to_database(uri, library=library, for_write=for_write, **kwargs)
        if system_type == "nosql":
            library = kwargs.pop("library", None)
            return connect_to_nosql(uri, library=library, for_write=for_write, **kwargs)
        raise ValueError(f"Unsupported system_type: {system_type}")

    def get_filesystem_connection(self, resource_name: str, for_write: bool=False, **kwargs):
        """
        Returns a FileSystem class for a given external filesystem, like s3, abs, gcs, etc..

        :param resource_name:   Which filesystem to access.  This can be the name of an input or output resource, or a
                                URI providing connection details.
        :param for_write:       True to enable write operations.
        :param kwargs:      Additional arguments, including...
                bucket = named bucket within the filesystem provider/engine.
        """
        return self._get_system_connection("filesystem", resource_name, for_write, **kwargs)

    def get_sql_connection(self, resource_name: str, for_write: bool=False, **kwargs):
        """
        Returns a DBI-compatible connection object which can be used to communicate with a relational database.

        :param resource_name:   Which database to access.  This can be the name of an input or output resource, or a
                                URI referencing an external data source.
        :param for_write:       True to enable write operations.
        """
        return self._get_system_connection("sql", resource_name, for_write, **kwargs)

    def get_nosql_connection(self, resource_name: str, for_write: bool=False, **kwargs):
        """
        Returns a connection to a NoSQL database, as an interface that abstracts the functionality of a range of
        popular NoSQL engines.

        :param resource_name:   Which database to access.  This can be the name of an input or output resource, or a
                                URI referencing an external data source.
        :param for_write:       True to enable write operations.
        """
        return self._get_system_connection("nosql", resource_name, for_write, **kwargs)

    def log_param(self, key: str, value):
        """
        Log a supplied parameter.
        """
        use_mlflow = self._mlflow_opts.get("params") or self._mlflow_all
        if use_mlflow:
            mlflow.log_param(key, value)
            return
        row = {
            "uri": self.get_uri(),
            "run_id": self._spec.get("run_id") or "",
            "user": self._spec.get("user") or "",
            "time": time.time(),
            "name": key,
            "value": value
        }
        df = pandas.DataFrame(data=[row])
        self.write_resource(PARAMS_ALIAS, df, append=True)

    def log_params(self, params: dict):
        """
        Log multiple parameters (if enabled).
        """
        for k, v in params.items():
            self.log_param(k, v)

    def log_metric(self, metric_name: str, metric_value: float, step: int=0) -> None:
        """
        Save a piece of trackable information about this run.
        """
        if self._mlflow_metrics:
            if mlflow:
                mlflow.log_metric(metric_name, metric_value, step)
            return
        if self._capturing_output(METRICS_ALIAS):
            # captured metrics only return a name and value
            if "capture" not in self._spec:
                self._spec["capture"] = {}
            if METRICS_ALIAS not in self._spec["capture"]:
                self._spec["capture"][METRICS_ALIAS] = {}
            self._spec["capture"][METRICS_ALIAS][metric_name] = metric_value
        else:
            # full metrics output
            row = {
                "uri": self.get_uri(),
                "run_id": self._spec.get("run_id") or "",
                "user": self._spec.get("user") or "",
                "time": time.time(),
                "name": metric_name,
                "value": metric_value,
                "step": step
            }
            df = pandas.DataFrame(data=[row])
            self.write_resource(METRICS_ALIAS, df, append=True)

    def log_metrics(self, metrics: dict, step: int=0) -> None:
        """
        Save multiple metrics.
        """
        for k, v in (metrics or {}).items():
            self.log_metric(k, v, step)

    def get_metrics(self, metric_name: str = None, uri: str = None, time_range: (list, tuple) = None, limit: int = None):
        """
        Retrieve the values of previously stored metrics as a dataframe.  Implementations might limit which metrics are returned
        based on security policies.  Sort order should be by time, descending.

        :param metric_name:     Limits results to a particular named metric.
        :param uri:             Limits results to a particular project, pipeline or model.
        :param time_range:      Limits results to a time range.  Supply two numeric values (or None), for start and
                                end of the time range, respectively.
        :param limit:           Maximum number of results to return.
        :returns:   An iterable of records (type 'Metric') describing each metric.
        """
        use_mlflow = self._mlflow_opts.get("metrics") or self._mlflow_all
        if use_mlflow:
            # TODO to support this would require creating a client, using the run ID to look them up, etc.
            raise DSLibraryException("Access to metrics in mlflow is not supported at this time.")
        # load metrics
        df = self.load_dataframe(METRICS_ALIAS)
        # filter by criteria, sort most recent first, limit to requested number of samples
        if time_range:
            df = df[time_range[0] <= df.time < time_range[1]]
        if uri:
            df = df[df.uri == uri]
        if metric_name:
            df = df[df.name == metric_name]
        df.sort_values(by=["time"], ascending=False)
        if limit:
            df = df.head(limit)
        return df

    def get_last_metric(self, metric_name: str):
        """
        Get the most recently saved metric of the given name for the current model.
        """
        metrics = self.get_metrics(metric_name, uri=self.get_uri(), limit=1)
        if metrics.shape[0]:
            row = metrics.iloc[0]
            return DSLibrary.Metric(**row)

    def log_text(self, text: str, artifact_file: str):
        """
        Log textual data as a run artifact / output.
        """
        self.write_resource(artifact_file, text)

    def log_dict(self, dictionary: dict, artifact_file: str):
        """
        Log JSON or YAML data.
        :param dictionary:
        :param artifact_file:
        """
        use_mlflow = self._mlflow_all
        if use_mlflow:
            mlflow.log_dict(dictionary, artifact_file)
        else:
            self.write_resource(artifact_file, content=dictionary)

    def log_artifact(self, local_path: str, artifact_path: str):
        """
        Send a file as a result from a model run.
        :param local_path:      Local file to send.
        :param artifact_path:   Filename for result file.
        """
        use_mlflow = self._mlflow_all
        if use_mlflow:
            mlflow.log_artifact(local_path, artifact_path)
        else:
            self.write_resource(artifact_path, from_filename=local_path)

    def log_artifacts(self, local_dir, artifact_path: str):
        """
        Send a whole folder.
        """
        for f in os.listdir(local_dir):
            self.log_artifact(os.path.join(local_dir, f), os.path.join(artifact_path, f))

    def next_scoring_request(self, timeout: float=None) -> (dict, None):
        """
        Use this method to iterate through multiple requests.  Returns a new {} on each request, or None if 'timeout'
        expired before any new requests were received.

        :param timeout:    How long to wait for a new request.
        """

    def scoring_response(self, score) -> None:
        """
        Reports a score from a scoring recipe.
        """

    def start_run(self):
        """
        Wraps the operation so that the actual start and end time can be known.  This is optional except when
        using MLFlow.
        """
        use_mlflow = self._mlflow_all
        if use_mlflow:
            mlflow.start_run()
            self._run_id = mlflow.active_run().info.run_id
        else:
            # generate a run ID
            self._run_id = str(uuid.uuid4())
            if mlflow:
                # capture some information to support active_run() when not using mlflow
                # TODO shouldn't we instead mock these mlflow classes, in case mlflow is not in the environment?
                from mlflow.entities import Run, RunData, RunInfo
                self._mlflow_run = Run(
                    RunInfo(
                        run_uuid=self._run_id, experiment_id=self._spec.get("uri"), user_id=self._spec.get("user"),
                        status="", start_time=time.time(), end_time=None, lifecycle_stage=""
                    ),
                    RunData()
                )
        inst = self
        class Wrap(object):
            def __enter__(self):
                return inst
            def __exit__(self, exc_type, exc_val, exc_tb):
                inst.end_run()
        return Wrap()

    def end_run(self):
        """
        Signals the end of the operation.
        """
        if self._mlflow_all:
            mlflow.end_run()

    def active_run(self):
        """
        For compatibility with MLFlow.
        """
        use_mlflow = self._mlflow_all
        if use_mlflow:
            return mlflow.active_run()
        return self._mlflow_run

    def install_packages(self, packages: list, verbose: bool=False):
        """
        Install packages.  Does a local 'pip install' by default but can be overridden to interact with different
        virtual environment managers.

        :param packages:  An iterable of package names, with optional qualifiers like "pandas==1.1.5"
        :param verbose:   By default we inhibit most output.
        """
        return install_packages(packages, verbose=verbose)

    def _write_content(self, opener, resource_name: str, content, append: bool=False, **kwargs):
        """
        Fully write various types of content.

        :param opener:      Method (i.e. self.open_resource() or similar) that will open a write stream.
        :param resource_name: Name of output resource where content should be stored.
        :param content:     Raw content for the data (various types), or a stream containing the data.
        :param append:      True to append, False to overwrite.
        """
        use_mlflow = self._mlflow_all
        # map resource name using supplied specs
        uri, kwargs = self._xlt_resource(resource_name, _mode="outputs", **kwargs)
        file_extension = find_url_extension(uri).strip(".")
        if not file_extension:
            self._default_format_for_aliases(resource_name, kwargs)
        format = kwargs.pop("format", None) or file_extension
        format_options = dict(kwargs.get("format_options") or {})
        # detect SQL target
        if format_options.get("sql_table"):
            # TODO formulate a bulk INSERT to write to the target table
            # TODO if 'append' is False, we'll have to first drop everything from the table
            # db_conn = self.get_sql_connection(resource_name)
            raise DSLibraryException("SQL write of columnar data is not yet implemented")
        if format_options.get("nosql_collection"):
            # TODO detect nosql (nosql_collection) and write data there
            raise DSLibraryException("NoSQL write of columnar data is not yet implemented")
        # we detect numpy arrays, series, etc., and save them as a dataframe with one column
        if isinstance(content, (pandas.Series, numpy.ndarray, pandas.Index)):
            series = content
            content = pandas.DataFrame()
            content[resource_name] = series
        # send (pandas) DataFrame as CSV/etc.
        df_format = format if format in PANDAS_WRITE_FORMATS else "csv"
        if hasattr(content, PANDAS_WRITE_FORMATS[df_format]):
            df_method = getattr(content, PANDAS_WRITE_FORMATS[df_format])
            # cancel append mode if file does not exist
            if append:
                try:
                    # verify file exists, capture column names
                    with opener(uri, 'rb') as f_r:
                        chk_args = {"nrows": 0}
                        if format == "json":
                            chk_args["lines"] = True
                        # apply write args to read format
                        for f in ["sep"]:
                            if f in format_options:
                                chk_args[f] = format_options[f]
                        df0 = PANDAS_READ_FORMATS[df_format](f_r, **chk_args)
                        # verify column names match
                        cols_l = set(df0.columns)
                        cols_r = set(content.columns)
                        if cols_l != cols_r:
                            raise DSLibraryException(f"Cannot append to {resource_name}, column names do not match: {', '.join(sorted(cols_l))} vs {', '.join(sorted(cols_r))}")
                except FileNotFoundError:
                    append = False
            mode = "w"
            if append:
                mode = "a"
                if format in ("csv", "tab"):
                    format_options.update({"header": False})
            if "index" not in format_options and format not in ("json",):
                format_options["index"] = False
            if format == "json":
                format_options.update({"orient": "records", "lines": True})
            with opener(uri, mode) as f_w:
                df_method(f_w, **format_options)
            return
        # save matplotlib figures as images
        elif hasattr(content, "savefig"):
            if append:
                raise ValueError("append is not supported for figures")
            if use_mlflow:
                mlflow.log_figure(content, uri)
                return
            buf = io.BytesIO()
            image_format = format or "png"
            content.savefig(buf, format=image_format)
            content = buf.getvalue()
        mode_base = 'a' if append else 'w'
        if isinstance(content, io.IOBase):
            # copy stream content
            with self.open_resource(resource_name, mode=mode_base + ('' if isinstance(content, io.TextIOBase) else 'b')) as f_w:
                while True:
                    chunk = content.read(5000000)
                    if not chunk:
                        break
                    f_w.write(chunk)
        elif isinstance(content, (str, bytes, bytearray)):
            # write content directly
            if use_mlflow:
                if not isinstance(content, str):
                    content = content.decode("utf-8")
                mlflow.log_text(content, uri)
            else:
                with self.open_resource(resource_name, mode=mode_base + ('' if isinstance(content, str) else 'b')) as f_w:
                    f_w.write(content)
        elif isinstance(content, (dict, list, tuple)):
            # write yaml or json
            if use_mlflow and isinstance(content, dict):
                mlflow.log_dict(content, uri)
                return
            if format in ("yaml", "yml"):
                with self.open_resource(resource_name, mode=mode_base) as f_w:
                    yaml.safe_dump(content, f_w)
                    f_w.write("\n")
            else:
                with self.open_resource(resource_name, mode=mode_base) as f_w:
                    json.dump(content, f_w)
                    f_w.write("\n")
        else:
            raise Exception(f"Unsupported content type: {type(content)}")

    def _write_file(self, opener, resource_name: str, filename: str, append: bool=False):
        """
        Copy a file or folder to an output resource.

        :param opener:      Method (i.e. self.open_resource() or similar) that will open a r/w stream.
        :param resource_name: Name of output resource to write to.
        :param filename:    Name of local file or folder to copy.
        :param append:      True to append, False to overwrite.
        """
        if os.path.isdir(filename):
            for f in os.listdir(filename):
                src = os.path.join(filename, f)
                dst = os.path.join(resource_name, f)
                self._write_file(opener, dst, src)
        else:
            with open(filename, 'rb') as f_r:
                with opener(resource_name, mode='ab' if append else 'wb') as f_w:
                    while True:
                        chunk = f_r.read(5000000)
                        if not chunk:
                            break
                        f_w.write(chunk)

    def read_resource(self, resource_name: str, mode: str='rb') -> (str, bytes):
        """
        Completely read the contents of a given named resource.  This is a convenience method for accessing small
        files.
        """
        with self.open_resource(resource_name, mode=mode) as f_r:
            return f_r.read()

    @staticmethod
    def _in_memory_append(a, b):
        if isinstance(a, str):
            return a + str(b)
        if isinstance(a, (bytes, bytearray)):
            return a + (b if isinstance(b, (bytes, bytearray)) else str(b).encode("utf-8"))
        if hasattr(a, "append"):
            return a.append(b)
        if isinstance(a, list):
            return a + list(b)
        if isinstance(a, dict) and isinstance(b, dict):
            out = dict(a)
            out.update(b)
            return out
        return b

    def _capturing_output(self, resource_name: str) -> bool:
        """
        Whether the given output is supposed to be captured.  This is part of the development/debugging feature that
        frees a caller from having to store i/o data in files to test a model.
        """
        return ((self._spec.get("outputs") or {}).get(resource_name) or {}).get("capture")

    def write_resource(self, resource_name: str, content=None, from_filename: str=None, append: bool=False, **kwargs):
        """
        Write various types of content or files to an output resource.
        :param resource_name:       Which output resource.
        :param content:             What to write (content):
            str, bytes, bytearray, file-like object:  Writes supplied content.
            pandas dataframe        Writes to CSV files, etc..
            matplotlib figure       Writes an image.
            dict, list              Writes JSON or YAML files.
        :param from_filename:       What to write (file or folder).
        :param append:              True to append, False to replace.
        """
        if content is not None:
            # content can be requested to be captured
            if self._capturing_output(resource_name):
                if "capture" not in self._spec:
                    self._spec["capture"] = {}
                if append:
                    prev = self._spec["capture"].get(resource_name)
                    if prev is not None:
                        content = self._in_memory_append(prev, content)
                self._spec["capture"][resource_name] = content
                return
            self._write_content(self.open_resource, resource_name, content=content, append=append, **kwargs)
        elif from_filename is not None:
            use_mlflow = self._mlflow_all
            if use_mlflow:
                mlflow.log_artifact(from_filename, resource_name)
                return
            self._write_file(self.open_resource, resource_name, filename=from_filename, append=append)
        else:
            raise ValueError("expected content or filename")

    def multi_part_upload(self, resource_name: str, chunk: (bytes, bytearray), upload_id: str=None, sequence: int=None, done: bool=False, **kwargs):
        """
        Manage a large upload.

        This is a placeholder for proper upload functionality for very large files, particularly to systems like S3 that
        do not support append.

        :param chunk:       Data to send.
        :param upload_id:   Unique identifier for the upload, returned from the first call.
        :param sequence:    Chunks can be delivered out of sequence, i.e. by separate processes.
        :param done:        Set this flag on the last chunk.
        :return:        An upload identifier.
        """
        # TODO this placeholder implementation just appends to a file.  This won't work for systems like s3.
        with self.open_resource(resource_name, mode='ab' if upload_id else 'wb', **kwargs) as f_w:
            f_w.write(chunk)
        return "-"

    def load_dataframe(self, resource_name: str, **kwargs):
        """
        Load a pandas dataframe from various types of input resources.

        Optional arguments:
            uri - path or URI to target
            format - csv, json, etc
            format_options - these are passed to pandas.read_csv(), etc.
            filesystem - name of a filesystem
            sql - SQL to execute
            sql_table - name of table to fully load
            nosql_collection - name of a NoSQL collection to fully load

        :param resource_name:   The name of an input resource, or of a local file, or the URI of an external resource.
        """
        # check for specifically supplied data
        supplied = (self._spec.get("data") or {}).get(resource_name)
        if supplied is not None and hasattr(supplied, "columns"):
            return supplied
        # map requested resource (resource_name) through input spec and override defaults (kwargs)
        uri, open_args = self._xlt_resource(resource_name, **kwargs)
        # detect SQL
        if open_args.get("sql") or open_args.get("sql_table"):
            sql = open_args.pop("sql", None)
            table = open_args.pop("sql_table", None)
            if table:
                sql = f"SELECT * from {sql_enquote_id(table)}"
            db_conn = self.get_sql_connection(resource_name, **open_args)
            try:
                c = db_conn.cursor()
                c.execute(sql)
                col_names = [col[0] for col in c.description]
                df = pandas.DataFrame(data=c, columns=col_names)
            finally:
                db_conn.close()
            return df
        # detect NoSQL
        if open_args.get("nosql_collection"):
            collection = open_args.pop("nosql_collection")
            db_conn = self.get_nosql_connection(resource_name, **open_args)
            rows = list(db_conn.query(collection))
            return pandas.DataFrame(rows)
        # TODO we should offer a validation option that checks the data types against the schema
        # format can be specified, or file extension will be used
        file_extension = find_url_extension(uri).strip(".")
        if not file_extension:
            self._default_format_for_aliases(resource_name, open_args)
        # options for read function can be given
        format = open_args.pop("format", None) or file_extension
        format_options = open_args.pop("format_options", None) or {}
        # various formats and their read function
        if format in PANDAS_READ_FORMATS:
            with self.open_resource(resource_name, mode='rb', **kwargs) as f_r:
                return PANDAS_READ_FORMATS[format](f_r, **format_options)
        # TODO support read from folder, i.e. folder of CSVs, parquet, etc.
        raise DSLibraryException(f"file type not supported: {format}")

    def _default_format_for_aliases(self, resource_name: str, options: dict):
        """
        Fill in default format for the built-in aliases.
        """
        if not options.get("format") and resource_name in (METRICS_ALIAS, PARAMS_ALIAS):
            options["format"] = "json"
            if "format_options" in options:
                format_options = options["format_options"]
            else:
                format_options = options["format_options"] = {}
            if "lines" not in format_options:
                format_options["lines"] = True

    def load_pickled_model(self, part: str=None):
        """
        Un-pickle a saved model from the default model binary, or from one of its named parts.
        """
        with self.open_model_binary(part=part, mode='rb') as f_r:
            return pickle.load(f_r)

    def save_pickled_model(self, model_object, part: str = None):
        """
        Use pickle to save a model.
        """
        with self.open_model_binary(part=part, mode='wb') as f_w:
            return pickle.dump(model_object, f_w)

    def read_run_data(self, resource_name: str, mode='rb') -> (str, bytes):
        """
        Completely read the contents of a given named 'run data' resource'.
        """
        with self.open_run_data(resource_name, mode=mode) as f_r:
            return f_r.read()

    def write_run_data(self, resource_name: str, content=None, filename: str=None, append: bool=False, **kwargs):
        """
        Fully write a named 'run data' object.
        """
        if content is not None:
            self._write_content(self.open_run_data, resource_name, content=content, append=append, **kwargs)
        elif filename is not None:
            self._write_file(self.open_run_data, resource_name, filename=filename, append=append)
        else:
            raise ValueError("expected content or filename")

    # TODO add remaining mlflow.tracking methods so that we can do this:
    #   mlflow = dslibrary.instance()


class DSLibraryException(Exception):
    """
    General errors.
    """


class DSLibraryCommunicationError(DSLibraryException):
    """
    Communication errors.
    """


LOCAL_SPECS_SCHEMA = {
    "comment": "Parameter values, and format/mapping information for inputs and outputs, as well as every other configurable run option is defined here.",
    "type": "object",
    "additionalProperties": False,
    "properties": {
        "uri": {"type": "string"},
        "user": {"type": "string"},
        "run_id": {"type": "string"},
        "entry_point": {"type": "string"},
        "parameters": {
            "comment": "Named parameter values for local testing.",
            "type": "object"
        },
        "inputs": {
            "comment": "Names of each input data source are mapped to appropriate local testing locations.",
            "type": "object",
            "patternProperties": {
                ".+": {
                    "comment": "Instructions for locating this particular named input to the model.  Parameters other than 'path' are passed through to the open() method, i.e. to fsspec.open().",
                    "type": "object",
                    "properties": {
                        "uri": {
                            "comment": "A URI to a remote resource can be given here, or a path (relative to the model root, or absolute) to a local file.  If omitted, blank or '.', the name of the input is used.",
                            "type": "string"
                        }
                    }
                }
            }
        },
        "outputs": {
            "comment": "Names of each output are mapped to appropriate local testing locations.  Same format as 'inputs'.",
            "type": "object"
        },
        "metrics": {
            "comment": "Where to send (or read) metrics.",
            "type": "object",
            "properties": {
                "uri": {
                    "comment": "Designates a columnar file or other data source where metrics will be stored",
                    "type": "string"
                }
            }
        },
        "mlflow": {
            "comment": "MLFlow-related settings",
            "type": "object",
            "properties": {
                "metrics": {"type": "boolean", "comment": "Uses MLFlow for tracking of metrics."},
                "all": {"type": "boolean", "comment": "Uses MLFlow for all available functionality."},
                "connection": {"comment": "uri, username, password"}
            }
        }
    }
}


PANDAS_READ_FORMATS = {
    "csv": pandas.read_csv,
    "tab": pandas.read_csv,
    "json": pandas.read_json,
    "parquet": pandas.read_parquet,
    "xlsx": pandas.read_excel,
    "xls": pandas.read_excel,
    "hdf": pandas.read_hdf,
    "fwf": pandas.read_fwf
}

PANDAS_WRITE_FORMATS = {
    "csv": "to_csv",
    "tab": "to_csv",
    "json": "to_json",
    "parquet": "to_parquet",
    "xlsx": "to_excel",
    "xls": "to_excel",
    "hdf": "to_hdf",
    "fwf": "to_fwf"
}



"""
Adapter to "mmlibrary".

The "mmlibrary" package can be installed locally to provide a default custom target.
"""
import io
import json
import inspect

from dslibrary.engine_intf import FileSystem
from dslibrary.front import DSLibraryException
from dslibrary.metadata import Metadata
from dslibrary.utils.file_utils import write_stream_with_read_on_close
from dslibrary import DSLibrary

try:
    import mmlibrary
except ImportError:
    mmlibrary = None


class DSLibraryViaMMLibrary(DSLibrary):
    def __init__(self, *args, _mm=None, **kwargs):
        """
        :param _mm:    You can reference a custom set of functions instead of the mmlibrary package.
        """
        self._mm = _mm or mmlibrary
        super(DSLibraryViaMMLibrary, self).__init__(*args, **kwargs)

    def get_metadata(self):
        return Metadata()

    def _find_method(self, *args, mandatory: bool=True):
        """
        Find a method that matches one of the names given in 'args'.
        """
        for method in args:
            if hasattr(self._mm, method):
                return getattr(self._mm, method)
        if mandatory:
            raise DSLibraryException(f"method not found: {args[0]}()")

    def get_parameters(self):
        if not self._mm:
            return {}
        get_arguments = self._find_method("get_arguments", "get_parameters", mandatory=False)
        if get_arguments:
            return self._mm.get_arguments()
        if hasattr(self._mm, "param_dictionary"):
            return self._mm.param_dictionary
        raise DSLibraryException(f"method not found: get_arguments()")

    def get_parameter(self, parameter_name: str, default=None):
        if not self._mm:
            return default
        method = self._find_method("get_parameter", "get_argument", "getArgument")
        try:
            return method(parameter_name)
        except (ValueError, KeyError):
            pass
        return default

    def _opener(self, path: str, mode: str, **kwargs) -> io.IOBase:
        if 'w' in mode:
            writer = self._find_method("save_binary_to_resource", "saveBinaryToResource")
            def finalize(fh):
                writer(path, fh.read())
            return write_stream_with_read_on_close(w_mode=mode, r_mode='rb', on_close=finalize)
        if "r" in mode:
            try:
                reader = self._find_method("get_binary_from_resource", "getBinaryFromResource")
                if mode == 'rb':
                    return io.BytesIO(reader(path))
                return io.StringIO(reader(path).decode('utf-8'))
            except ValueError:
                raise FileNotFoundError(f"not found: {path}")
        raise DSLibraryException(f"Unsupported mode: {mode}")

    def get_filesystem_connection(self, resource_name: str, for_write: bool=False, **kwargs):
        open_method = self._find_method("open_stream")
        if not open_method:
            raise DSLibraryException(f"No filesystem engine access available, requested={resource_name}")
        bucket = kwargs.get("bucket")
        file_source = f"{resource_name}:{bucket}" if bucket else resource_name
        class MMFS(FileSystem):
            def ls(self, path, detail=True, **kwargs):
                raise NotImplementedError()
            def rm(self, path, recursive=False, maxdepth=None):
                raise NotImplementedError()
            def mkdir(self, path, create_parents=True, **kwargs):
                raise NotImplementedError()
            def stat(self, path, **kwargs):
                raise NotImplementedError()
            def open(self, path, mode="rb", block_size=None, cache_options=None, **kwargs):
                return open_method(path, file_source=file_source, mode=mode)
        return MMFS()

    def open_model_binary(self, part: str=None, mode: str='rb') -> io.RawIOBase:
        """
        The model-binary is just one file, but we can support 'part' by assuming it is a ZIP file.
        """
        if part:
            import zipfile
            if mode in ("w", "wb", "a", "ab"):
                # we can't write parts because each call to mmlibrary.new_version() creates a new trained model
                raise ValueError(f"Unsupported mode for model binary with parts: {mode}")
            with self.open_model_binary() as f_r:
                with zipfile.ZipFile(f_r) as zf:
                    if mode == "r":
                        return io.StringIO(zf.read(part).decode("utf-8"))
                    if mode == "rb":
                        return io.BytesIO(zf.read(part))
        if "r" in mode:
            get_model = self._find_method("get_model", "getModel")
            return open(get_model(), mode)
        if "w" in mode:
            wr_model = self._find_method("new_version", "newVersion")
            return write_stream_with_read_on_close(w_mode=mode, r_mode='rb', on_close=lambda f_r: wr_model(f_r.read()))

    def log_metric(self, metric_name: str, metric_value: float, step: int=0) -> None:
        """
        MMlibrary has its own KPI logging method.
        """
        if self._mlflow_metrics:
            return super(DSLibraryViaMMLibrary, self).log_metric(metric_name, metric_value, step)
        self._mm.save_kpi(metric_name, metric_value)

    def open_run_data(self, filename: str, mode: str='rb') -> io.IOBase:
        """
        The method provided by mmlibrary to save 'run data' only stores a single chunk of binary data.  We
        store JSON in that data, and translate calls to read and write files such that they read and write
        elements of that JSON object.
        """
        reader = self._find_method("get_temporary_data")
        if "r" in mode:
            data = json.loads(reader() or b'{}')
            if filename not in data:
                raise FileNotFoundError(f"not found: {filename}")
            if mode == 'r':
                return io.StringIO(data[filename])
            return io.BytesIO((data[filename]).encode("utf-8"))
        def finalize(fh):
            data = json.loads(reader() or b'{}')
            if "a" in mode:
                if filename not in data:
                    data[filename] = ""
                data[filename] += fh.read()
            else:
                data[filename] = fh.read()
            writer = self._find_method("save_temporary_data")
            writer(json.dumps(data).encode('utf-8'))
        return write_stream_with_read_on_close(w_mode=mode, r_mode='r', on_close=finalize)

    def get_next_scoring_request(self, timeout: float=None) -> (dict, None):
        raise DSLibraryException("not implemented in mmlibrary, use get_parameter() for each fields of a single scoring request")

    def scoring_response(self, score):
        method = self._find_method("return_score", "returnScore")
        method(score)

    def get_sql_connection(self, resource_name: str, for_write: bool=False, database: str=None, **kwargs):
        method = self._find_method("get_db_connection", "getDBConnection")
        sig = inspect.signature(method)
        params = {}
        if database:
            if "database" in sig.parameters:
                params["database"] = database
            else:
                raise DSLibraryException("The 'database' parameter is not supported by this implementation of get_db_connection()")
        return method(resource_name, **params)

    def install_packages(self, packages: list, verbose: bool=False):
        method = self._find_method("install_packages")
        sig = inspect.signature(method)
        if method:
            return method(packages)
        else:
            return super(DSLibraryViaMMLibrary, self).install_packages(packages, verbose)



"""
Package management - interactions with pip, virtual environments.
"""
import re
import subprocess
import sys
import tempfile

PTN_EXPR = re.compile(r'^\s*([A-Za-z0-9\-._]+)(\[[A-Za-z0-9.\-_]+])?(\s*(=|==|>|>=|<|<=)\s*([0-9.\-a-zA-Z]+)\s*,?)*(\s+-[^\s]+)*\s*?$')


def current_packages():
    """
    Get the list of currently installed packages.
    """
    stdout = subprocess.check_output([sys.executable, "-m", "pip", "freeze"])
    out = {}
    for line in stdout.decode("utf-8").split("\n"):
        if "==" not in line:
            continue
        package, version = line.split("==", maxsplit=1)
        out[package] = version
    return out


def parse_package_expr(line: str):
    """
    Verify the format of a package expression is correct.
    """
    if not line or not line.strip() or line.strip().startswith("#"):
        return
    m = PTN_EXPR.match(line)
    if not m:
        raise ValueError(f"Invalid package specification: {line}")
    package = m.group(1)
    options = m.group(2)
    versions = m.group(3)
    flags = m.group(6)
    # TODO compiled test to see if a package version matches
    # TODO work out the ideal version to install (latest or something else), or rather the whole set of options
    return line


def install_packages(packages: list, verbose: bool=False):
    """
    Install packages.
    """
    for line in packages:
        parse_package_expr(line)
    with tempfile.NamedTemporaryFile(suffix=".requirements.txt", mode='w') as f_tmp:
        f_tmp.write("\n".join(packages))
        f_tmp.flush()
        subprocess.check_call([sys.executable, "-m", "pip", "install", *([] if verbose else ["-q"]), "--disable-pip-version-check", "-r", f_tmp.name])



import pathlib
from setuptools import setup

HERE = pathlib.Path(__file__).parent
README = (HERE / "README.md").read_text()

setup(
    name="dslibrary",
    version="0.0.11",
    description="Data Science Framework & Abstractions",
    long_description=README,
    long_description_content_type="text/markdown",
    #url="n/a",
    #author="n/a",
    #author_email="n/a",
    license="Apache",
    classifiers=[
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3"
    ],
    packages=["dslibrary"],
    include_package_data=True,
    install_requires=["jsonschema", "pyyaml", "pandas", "pexpect"],
    extras_require={
        "all": [
            "fsspec", "boto3", "s3fs", "gcsfs", "adlfs",    # file transfer to the main CSPs
            "psycopg2-binary", "PyMySQL", "sqlite3",        # a few SQL databases
            "openpyxl"                                      # r/w excel files
        ]
    }
    #entry_points={
    #    "console_scripts": [
    #        "dslibrary=dslibrary.__main__:main",
    #    ]
    #}
)


"""
File/stream related tools.
"""
import io
import os
import tempfile
import typing
import urllib.request

from dslibrary.engine_intf import FileSystemLocal, FileSystemReadOnly

try:
    import fsspec
except ImportError:
    fsspec = None


class FileOpener(object):
    """
    Opens read/write streams on a wide range of URIs.
    """
    def __init__(self, root_folder: str=None):
        """
        :param root_folder:   Default folder, in case relative local paths are given.
        """
        self._root = os.path.expanduser(root_folder or ".")

    def http_open(self, path: str, mode: str, **kwargs):
        """
        HTTP and FTP can be handled with the built-in urllib.

        :param path:        URI specifying an HTTP or FTP location.
        :param mode:        'r' and 'rb' are supported.
        :param kwargs:      'headers' can be specified.
        :return:        A stream, if the supplied arguments were supported, or None.
        """
        if path.startswith("https://") or path.startswith("http://") or path.startswith("ftp://"):
            if mode == "rb":
                request = urllib.request.Request(url=path, headers=kwargs.get("headers", None))
                return urllib.request.urlopen(request, **kwargs)
            if mode == "r":
                byte_stream = self.http_open(path, mode, **kwargs)
                return io.TextIOWrapper(byte_stream)
            if mode in ("w", "wb"):
                # data can be sent to a REST service
                #  - set method to POST or PUT
                #  - set any needed authentication headers in 'headers'
                #  - path is URL plus any required arguments in the query string
                request = urllib.request.Request(url=path, headers=kwargs.get("headers", None), method=kwargs.get("method", "POST"))
                return write_stream_with_read_on_close(mode, r_mode='rb', on_close=lambda stream: urllib.request.urlopen(request, data=stream))
            raise ValueError(f"Append mode is not supported for {path}")

    def uri_open(self, path: str, mode: str, **kwargs):
        """
        Open specific/exotic URIs, like s3, abfs, adl, gs, etc..

        :param path:        The URI, usually of the form PROTO://BUCKET/PATH.
        :param mode:        'r', rb', 'w', etc.
        :param kwargs:      Additional options passed through to the target library (i.e. fsspec).
        :return:        A file stream, if a handler was found.
        """
        if "://" not in path:
            return
        if not fsspec:
            raise ImportWarning(f"The 'fsspec' package is required to open {path}")
        return fsspec.open(path, mode=mode, **kwargs)

    def local_open(self, path: str, mode: str, **kwargs):
        """
        Open local files.
        :param path:    A relative path, relative to self._root, or an absolute path.
        :param mode:    'r', 'rb', 'w', 'wb', 'a', 'ab'.
        :param kwargs:  No additional options are recognized.
        :return:    A file stream.
        """
        full_path = os.path.join(self._root, path)
        return open(full_path, mode)

    def open(self, path: str, mode: str, **kwargs):
        """
        Open files.

        :param path:    Path to file, or URI of file.
        :param mode:    Open mode (r, rb, w, wb, a, ab)
        :param kwargs:  Additional arguments to customize details of the operation.
        :return:    File-like object.
        """
        # some really simple external file access can be handled without fsspec
        stream = self.http_open(path, mode, **kwargs)
        if stream:
            return stream
        # URIs go through fsspec
        stream = self.uri_open(path, mode, **kwargs)
        if stream:
            return stream
        # local files
        return self.local_open(path, mode, **kwargs)


def write_stream_with_read_on_close(w_mode: str, r_mode: str, on_close: typing.Callable):
    """
    Open a write stream that saves data in a temporary file and sends a read stream for that data to a supplied method.
    :param w_mode:          'w' or 'wb'.
    :param r_mode:          'r' or 'rb'.
    :param on_close:        Method called with read stream
    :return:
    """
    fh = tempfile.NamedTemporaryFile(mode=w_mode, delete=False)
    def closer():
        if not fh.closed:
            fh.flush()
        with open(fh.name, r_mode) as f_r:
            on_close(f_r)
        os.remove(fh.name)
    fh.close = closer
    return fh


def connect_to_filesystem(uri: str, for_write: bool=False, **kwargs):
    """
    Generate an open(filename, mode) method based on a URI and credentials.
    :param uri:             A partial URI like "s3://bucket", or a local path.
    :param for_write:       Whether to allow write access.
    :param kwargs:          Additional arguments, namely credentials.
    :return:            A FileSystem instance that lets you work with files in the remote filesystem.
    """
    if "://" not in uri:
        # uri can just be a local path
        fs = FileSystemLocal(uri)
    else:
        # otherwise we assume the usual cloud format: protocol://bucket
        protocol, bucket = uri.strip("/").split("://", maxsplit=1)
        bucket = bucket.split("/")[0]
        if ":" in bucket:
            raise ValueError(f"Invalid URI for external filesystem: {uri}")
        if not fsspec:
            raise ImportWarning(f"fsspec not installed, can't open filesystem: {uri}")
        fs = fsspec.filesystem(protocol, bucket=bucket, **kwargs)
    if not for_write:
        return FileSystemReadOnly(fs)
    return fs


def find_url_extension(url: str) -> str:
    """
    Find the extension of a filename or URL.
    """
    url = url or ""
    if "#" in url:
        url = url.split("#")[0]
    if "?" in url:
        url = url.split("?")[0]
    return os.path.splitext(url)[1].lower().strip(".")
